function m = cnpkg_buildmodel(p, outSize)

% M = cnpkg_buildmodel(P) returns a CNS network model structure for a 3-D
% convolutional network which can be used as input to cns('init').  P is a
% small set of parameters that define the network, and OUTSIZE is the spatial
% size [Y*X*D] of the output level.  See the example in cnpkg_run for details.
%
% See also: cnpkg_run.

%***********************************************************************************************************************

n = numel(p.fCount); % Number of 'X' levels.

zx = [1, 4 + (0 : n - 2) * 3]; % CNS layer numbers for all levels.
zw = [0, 2 + (0 : n - 2) * 3]; % CNS layer numbers for all weight layers.
zb = [0, 3 + (0 : n - 2) * 3]; % CNS layer numbers for all bias layers.

m.package = 'cnpkg';

% Set up the input level.

m.layers{zx(1)}.z       = zx(1);       % For display only.
m.layers{zx(1)}.name    = 'x1';
m.layers{zx(1)}.type    = 'input';
m.layers{zx(1)}.stepNo  = 1;           % Step number for forward pass.
m.layers{zx(1)}.size{1} = p.fCount(1); % Sizes of spatial dimensions to be determined by cns_mapdim (below).

% Set up the hidden levels.

for i = 2 : n - 1

    m.layers{zx(i)}.z       = zx(i);              % For display only.
    m.layers{zx(i)}.name    = sprintf('x%u', i);
    m.layers{zx(i)}.type    = 'hidden';
    m.layers{zx(i)}.stepNo  = [i, 2 * n - i + 1]; % Step numbers for forward and backward passes.
    m.layers{zx(i)}.kernel  = {'', 'back'};
    m.layers{zx(i)}.zp      = zx(i - 1);          % Pointer to previous level (for forward pass).
    m.layers{zx(i)}.zw      = zw(i);              % Pointer to previous weight layer (for forward pass).
    m.layers{zx(i)}.zb      = zb(i);              % Pointer to previous bias layer (for forward pass).
    m.layers{zx(i)}.znw     = zw(i + 1);          % Pointer to next weight layer (for backward pass).
    m.layers{zx(i)}.zn      = zx(i + 1);          % Pointer to next level (for backward pass).
    m.layers{zx(i)}.size{1} = p.fCount(i);        % Sizes of spatial dimensions to be determined by cns_mapdim (below).

end

% Set up the output level.

m.layers{zx(n)}.z       = zx(n);             % For display only.
m.layers{zx(n)}.name    = sprintf('x%u', n);
m.layers{zx(n)}.type    = 'output';
m.layers{zx(n)}.stepNo  = [n, n + 1];        % Step numbers for forward and backward passes.
m.layers{zx(n)}.kernel  = {'', 'back'};
m.layers{zx(n)}.zp      = zx(n - 1);         % Pointer to previous level (for forward pass).
m.layers{zx(n)}.zw      = zw(n);             % Pointer to previous weight layer (for forward pass).
m.layers{zx(n)}.zb      = zb(n);             % Pointer to previous bias layer (for forward pass).
m.layers{zx(n)}.size{1} = p.fCount(n);       % Sizes of spatial dimensions to be determined by cns_mapdim (below).

% Set up weight and bias layers.

for i = 2 : n

    m.layers{zw(i)}.z       = zw(i);             % For display only.
    m.layers{zw(i)}.name    = sprintf('w%u', i);
    m.layers{zw(i)}.type    = 'weight';
    m.layers{zw(i)}.stepNo  = 2 * n;             % All weights are updated in parallel after the backward pass.
    m.layers{zw(i)}.zp      = zx(i - 1);         % Pointer to previous level (for update).
    m.layers{zw(i)}.zn      = zx(i);             % Pointer to next level (for update).
    m.layers{zw(i)}.eta     = p.eta(i);
    m.layers{zw(i)}.size{1} = p.fCount(i - 1);
    m.layers{zw(i)}.size{2} = p.fSize (i);
    m.layers{zw(i)}.size{3} = p.fSize (i);
    m.layers{zw(i)}.size{4} = p.fDepth(i);
    m.layers{zw(i)}.size{5} = p.fCount(i);

    m.layers{zb(i)}.z       = zb(i);             % For display only.
    m.layers{zb(i)}.name    = sprintf('b%u', i);
    m.layers{zb(i)}.type    = 'bias';
    m.layers{zb(i)}.stepNo  = 2 * n;             % All biases are updated in parallel after the backward pass.
    m.layers{zb(i)}.zn      = zx(i);             % Pointer to next level (for update).
    m.layers{zb(i)}.eta     = p.eta(i);
    m.layers{zb(i)}.size{1} = p.fCount(i);

end

% Set the spatial size of the output level and assign each output cell to a position in the real-valued, common
% coordinate system.

m = cns_mapdim(m, zx(n), 2, 'pixels', outSize(1));
m = cns_mapdim(m, zx(n), 3, 'pixels', outSize(2));
m = cns_mapdim(m, zx(n), 4, 'pixels', outSize(3));

% Now grow the network downwards from output to input, creating each level with the appropriate size, and the
% appropriate arrangement of its cells in the common coordinate system, such that it will generate the level above via
% valid convolution.  Each level will be larger than the one above it.

for i = n - 1 : -1 : 1

    m = cns_mapdim(m, zx(i), 2, 'int-td', zx(i + 1), p.fSize (i + 1), 1);
    m = cns_mapdim(m, zx(i), 3, 'int-td', zx(i + 1), p.fSize (i + 1), 1);
    m = cns_mapdim(m, zx(i), 4, 'int-td', zx(i + 1), p.fDepth(i + 1), 1);

end

% This flag turns off double-buffering in CNS, which is not needed for this model.  See the CNS manual.

m.independent = true;

% CNS doesn't need these, but we'll store them in the model structure for our own use.

m.zx = zx;
m.zw = zw;
m.zb = zb;

return;