% cnpkg_run is an example script that shows how to use cnpkg.  We train a
% network to perform denoising using a 2-D image.  Note that cnpkg supports 3-D
% networks; here the "depth" dimension has size 1.
%
% See also: cnpkg_buildmodel.

%***********************************************************************************************************************

p = struct;                  % Network parameters.  We have 1 input level, 2 hidden levels, and 1 output level.
p.fCount = [3 10 10 3];      % Number of feature maps in each level.
p.fSize  = [0 7 7 1];        % Size of weights in Y and X for each level.
p.fDepth = [0 1 1 1];        % Size of weights in D (depth) for each level.
p.eta    = [0 1 1 1] * 1e-2; % Learning rate for weights and biases for each level.

miniOutSize = [3 3 1];       % Spatial size of output level for training [Y*X*D].

im = imread('autumn.tif');

if ~exist('nEpoch', 'var'), nEpoch = 100  ; end
if ~exist('nIter' , 'var'), nIter  = 50000; end

% Prepare training set.

im = single(im) / single(intmax(class(im)));

input = im + single(0.1 * randn(size(im))); % Add noise.

border = sum(p.fSize(2 : end) - 1) / 2; % Input is larger than output due to edge loss from the convolutions.

label     = im   (border + 1 : end - border, border + 1 : end - border, :);
dispInput = input(border + 1 : end - border, border + 1 : end - border, :);

dispInput = min(max(dispInput, 0), 1);

% Create two models, one for training and one for testing.

m1 = cnpkg_buildmodel(p, miniOutSize);
m1.input = permute(input, [3 1 2]); % Convert to [F*Y*X*D].
m1.label = permute(label, [3 1 2]); % Convert to [F*Y*X*D].

m2 = cnpkg_buildmodel(p, [size(label, 1), size(label, 2), 1]);
m2.input = permute(input, [3 1 2]);
m2.label = permute(label, [3 1 2]);
m2.batch = [0 0 0];

% Initialize weights.

for i = 2 : numel(p.fCount)
    siz = cell2mat(m1.layers{m1.zw(i)}.size);
    m1.layers{m1.zw(i)}.val = single(0.5 * randn(siz) / sqrt(prod(siz(1 : 4))));
    m1.layers{m1.zb(i)}.val = single(0.5 * randn(siz(5), 1));
end

% Loop over epochs.

eloss = zeros(1, nEpoch);

fprintf('training (%u epochs, %u iterations per epoch):\n', nEpoch, nIter);

for epoch = 1 : nEpoch

    % Randomly select minibatches.

    m1.batch = zeros(nIter, 3);
    m1.batch(:, 1) = randi(size(label, 1) - miniOutSize(1) + 1, [nIter 1]) - 1;
    m1.batch(:, 2) = randi(size(label, 2) - miniOutSize(2) + 1, [nIter 1]) - 1;
    m1.batch(:, 3) = 0;

    % Iterate the training model once per minibatch.  Each iteration consists of a forward pass, a backward pass, and
    % a weight update.

    t1 = clock;

    cns('init', m1, 'gpu', 'mean');
    cns('run', nIter);

    % Retrieve weights, transfer them to testing model.

    for i = 2 : numel(p.fCount)
        m1.layers{m1.zw(i)}.val = cns('get', m1.zw(i), 'val');
        m1.layers{m1.zb(i)}.val = cns('get', m1.zb(i), 'val');
        m2.layers{m2.zw(i)}.val = m1.layers{m1.zw(i)}.val;
        m2.layers{m2.zb(i)}.val = m1.layers{m1.zb(i)}.val;
    end

    t1 = etime(clock, t1);

    % Iterate the testing model once and plot status.

    t2 = clock;

    cns('init', m2, 'gpu', 'mean');
    cns('step', 1, m2.layers{m2.zx(end)}.stepNo(1)); % Don't need to backpropagate, etc.

    output = permute(cns('get', m2.zx(end), 'val'), [2 3 1]);

    t2 = etime(clock, t2);

    loss = mean(0.5 * (label - output) .^ 2, 3);
    output = min(max(output, 0), 1);

    eloss(epoch) = mean(loss(:));

    fprintf('completed epoch #%u of %u (train: %f sec, test: %f sec)\n', epoch, nEpoch, t1, t2);

    figure(1); clf;
    plot(eloss(1 : epoch)); title('loss'); xlabel('epoch');

    figure(2); clf; ax = [];
    ax(1) = subplot(221); image  (label    ); title('clean image'   ); axis image off;
    ax(2) = subplot(222); image  (dispInput); title('noisy image'   ); axis image off;
    ax(3) = subplot(223); imagesc(loss     ); title('loss image'    ); axis image off; colorbar;
    ax(4) = subplot(224); image  (output   ); title('denoised image'); axis image off;
    linkaxes(ax);

end

% Release GPU resources.

cns('done');
